#include "header\gba.h"
#include "header\bg.h"
#include "header\timer.h"
#include "header\dma.h"
#include "header\irq.h"
#include "header\sound.h"
#include "header\sprite.h"
#include "header\blend.h"

#include "header\font\BGFont.h"
//#include "header\font\BitMapFont.h"

#include "lib.h"
#include "syscall.h"
#include "resource.h"

#include <stdlib.h>

//Mode0
//BG0 
//BG3 Xe[W

//---------------------------------------------------------------------------
#define STATE_INIT 					1
#define STATE_RESET					2
#define STATE_MOVE					3
#define STATE_JUMP					4
#define STATE_FALL					5
#define STATE_END					6

#define STAGE_CX					30
#define STAGE_CY					32
#define STAGE_INIT_WALL_LINE		20
#define STAGE_SCROLL_LINE			32
#define STAGE_GANERATE_INTERVAL		5

#define PLAYER_INIT_JUMP_A			12
#define PLAYER_INIT_JUMP_B			10
#define PLAYER_INIT_GRAVITY			2
#define PLAYER_MAX_GRAVITY			15

#define BG_TILE_EMPTY				0
#define BG_TILE_WALL				2

//---------------------------------------------------------------------------
typedef struct {
	u8  bg[STAGE_CX][STAGE_CY];
	u8  hofs;
	u8  vofs;
	s16 genarateLine;
} ST_STAGE;

typedef struct {
	u8  x;
	u8  y;
//	u8  wx;
	u8  wy;
	u8  cx;
	u8  cy;
	s16 moveX;
	s16 moveY;
	u16 sprNo;
	u16 sprTile;
	u16 sprSize;
	u16 sprForm;
} ST_PLAYER;

//---------------------------------------------------------------------------
//lib.c
extern ST_TIMER Timer;

//---------------------------------------------------------------------------
ST_BG     BG0;
ST_BG     BG3;
ST_STAGE  Stage;
ST_PLAYER Player;

u8  State;
//---------------------------------------------------------------------------
void StateInit();
void StateReset();
void StateMove();
void StateJump();
void StateFall();

void InitWS();
void InitBG();
void InitSpr();
void InitKey();
void IRQUserHandler();

void InitStage();
void InitPlayer();

void GenerateRandomWallLine();

void MovePlayerRow();
void MovePlayerRowSideWall();
void MovePlayerJump();
void MovePlayerFall();

s16  GetReviseGravity();

void SetBGScroll();

void DrawBGStageAll();
void DrawSprPlayer();

bool ChkTouchWall(u8 x, u8 y, u8 cx, u8 cy);
bool ChkFall();
bool ChkInputJump();

bool ChkScrollLine();
void ScrollLine();

//---------------------------------------------------------------------------
int main()
{
	State = STATE_INIT;

	for(;;)
	{
		WaitForVsync();

		switch(State)
		{
			case STATE_INIT:  StateInit();  break;
			case STATE_RESET: StateReset(); break;
			case STATE_MOVE:  StateMove();  break;
			case STATE_JUMP:  StateJump();  break;
			case STATE_FALL:  StateFall();  break;

			default:
				DebugMessage((u8*)"Err: Main State");
				break;
		}
	}

	return 0;
}
//---------------------------------------------------------------------------
void StateInit()
{
	InitWS();
	InitBG();
	InitSpr();
	InitKey();

	SetMode(MODE_0 | BG0_ENABLE | BG2_ENABLE | BG3_ENABLE | OBJ_ENABLE | OBJ_MAP_2D);
	State = STATE_RESET;
}
//---------------------------------------------------------------------------
void InitWS()
{
	REG_WSCNT = CST_ROM0_1ST_3WAIT | CST_ROM0_2ND_1WAIT;
}
//---------------------------------------------------------------------------
void InitBG()
{
	//BG0 
	//BG3 Xe[W

	BG0.chr = 1;
	BG0.map = 0;
	BG0.col = 0;

	BG3.chr = 2;
	BG3.map = 1;
	BG3.col = 1;

	REG_BG0CNT = (BG_SIZEA_256_256     | BG_COLOR_16         |
				  BG_CHARBASE(BG0.chr) | BG_MAPBASE(BG0.map) | 0);

	REG_BG3CNT = (BG_SIZEA_256_256     | BG_COLOR_16         |
				  BG_CHARBASE(BG3.chr) | BG_MAPBASE(BG3.map) | 0);

	BGSetPal16  (BG0.col, (u16*)&BGFont_palette);
	BGSetChrData(BG0.chr, (u16*)&BGFont_data, BGFont_DATA_SIZE);

	BGSetPal16  (BG3.col, (u16*)&game_bg_pal_bin);
	BGSetChrData(BG3.chr, (u16*)&game_bg_img_bin, GAME_BG_IMG_SIZE);
}
//---------------------------------------------------------------------------
void InitSpr()
{
	SpriteInit();

	SpriteSetPal16((u16*)&game_spr_pal_bin);
	SpriteSetData ((u16*)&game_spr_img_bin, GAME_SPR_IMG_SIZE);
}
//---------------------------------------------------------------------------
void InitKey()
{
	KeyInit();

	//L[͗p̃^C}ݒ
	REG_INTERUPT = (u32)IRQUserHandler;
	REG_IE       = IRQ_BIT_TIMER0;
	REG_IME      = IRQ_MASTER_ON;

	REG_TM0D     = 65536 - 2621; // (((16*1024*1024) / 64) * 10) / 1000
	REG_TM0CNT   = TM_ENABLE | TM_FREQ_PER_64 | TM_USEIRQ;
}
//---------------------------------------------------------------------------
void IRQUserHandler()
{
	Timer.clockTick += TIMER_INTMS;
	REG_IF = IRQ_BIT_TIMER0;
}
//---------------------------------------------------------------------------
void StateReset()
{
	InitStage();
	InitPlayer();

	SetBGScroll();

	DrawBGStageAll();
	DrawSprPlayer();

	State = STATE_MOVE;
}
//---------------------------------------------------------------------------
void InitStage()
{
	ST_STAGE* p = &Stage;

	s16 x, y;
	for(y=0; y<STAGE_CY; y++)
	{
		for(x=0; x<STAGE_CX; x++)
		{
			p->bg[x][y] = BG_TILE_EMPTY;
		}
	}

	for(x=0; x<STAGE_CX; x++)
	{
		p->bg[x][STAGE_INIT_WALL_LINE-1] = BG_TILE_WALL;
	}
	for(y=0; y<STAGE_CY; y++)
	{
		p->bg[0][y]          = BG_TILE_WALL;
		p->bg[STAGE_CX-1][y] = BG_TILE_WALL;
	}

	p->genarateLine = 14;
	s16 i;
	for(i=0; i<3; i++)
	{
		GenerateRandomWallLine();
	}

	//Scroll
	p->hofs = 0;
	p->vofs = 0;
}
//---------------------------------------------------------------------------
void InitPlayer()
{
	ST_PLAYER* p = &Player;

	p->x        =  1 * BG_CX;	//\WiXN[␳j
	p->y        = 18 * BG_CY;
	p->wy       =  p->y;		//΍W
	p->cx       =  8 - 1;
	p->cy       =  8 - 1;
	p->moveX    =  1;
	p->moveY    =  0;
	p->sprNo    =  0;
	p->sprTile  =  0;
	p->sprSize  = SP_SIZE_8;
	p->sprForm  = SP_SQUARE;

	SpriteSetSize(p->sprNo, p->sprSize, p->sprForm, SP_COLOR_16);
	SpriteSetChr (p->sprNo, p->sprTile);
}
//---------------------------------------------------------------------------
void DrawSprPlayer()
{
	ST_PLAYER* p = &Player;
	SpriteMove(p->sprNo, p->x, p->y);
}
//---------------------------------------------------------------------------
void DrawBGStageAll()
{
	ST_STAGE* p = &Stage;

	s16 x, y;
	for(y=0; y<STAGE_CY; y++)
	{
		for(x=0; x<STAGE_CX; x++)
		{
			BGDrawChar(BG3.map, x, y, BG3.col, p->bg[x][y]);
		}
	}
}
//---------------------------------------------------------------------------
void SetBGScroll()
{
	ST_STAGE* p = &Stage;

	REG_BG3HOFS = p->hofs;
	REG_BG3VOFS = p->vofs;
}
//---------------------------------------------------------------------------
void StateMove()
{
	if( ChkInputJump() == TRUE )
	{
		State = STATE_JUMP;
		return;
	}
	if( ChkFall() == TRUE )
	{
		State = STATE_FALL;
		return;
	}

	MovePlayerRow();
	DrawSprPlayer();
}
//---------------------------------------------------------------------------
bool ChkTouchWall(u8 x, u8 y, u8 cx, u8 cy)
{
	ST_STAGE* p = &Stage;

	u8 x1 = (u8)(x     ) / BG_CX;
	u8 y1 = (u8)(y     ) / BG_CY;
	u8 x2 = (u8)(x + cx) / BG_CX;
	u8 y2 = (u8)(y + cy) / BG_CY;

	if(p->bg[x1][y1] == BG_TILE_WALL) return TRUE;
	if(p->bg[x1][y2] == BG_TILE_WALL) return TRUE;
	if(p->bg[x2][y1] == BG_TILE_WALL) return TRUE;
	if(p->bg[x2][y2] == BG_TILE_WALL) return TRUE;

	return FALSE;
}
//---------------------------------------------------------------------------
void MovePlayerRow()
{
	ST_PLAYER* p = &Player;

	if( ChkTouchWall(p->x + p->moveX, p->wy, p->cx, p->cy) == TRUE )
	{
		p->moveX *= -1;
	}

	p->x += p->moveX;
}
//---------------------------------------------------------------------------
void MovePlayerRowSideWall()
{
	ST_PLAYER* p = &Player;

	if(   p->x + p->moveX         < BG_CX
	   || p->x + p->cx + p->moveX > BG_CX * (STAGE_CX-1) )
	{
		p->moveX *= -1;
	}

	p->x += p->moveX;
}
//---------------------------------------------------------------------------
bool ChkInputJump()
{
	if( KeyRepeat() == TRUE )		//keyGet֐ĂԑOɕKsȂ
	{
		return FALSE;
	}

	u8 key = KeyGet();
	if( !(key & KEY_A) && !(key & KEY_B) )
	{
		return FALSE;
	}

	ST_PLAYER* p = &Player;

	if(key & KEY_A) p->moveY = PLAYER_INIT_JUMP_A;
	if(key & KEY_B) p->moveY = PLAYER_INIT_JUMP_B;

	return TRUE;
}
//---------------------------------------------------------------------------
bool ChkFall()
{
	ST_PLAYER* p = &Player;

	if(   ChkTouchWall(p->x, p->wy + 0, p->cx, p->cy) == FALSE
	   && ChkTouchWall(p->x, p->wy + 1, p->cx, p->cy) == TRUE )
	{
		return FALSE;
	}

	return TRUE;
}
//---------------------------------------------------------------------------
void StateJump()
{
	ST_PLAYER* p = &Player;
	if(p->moveY <= 0)
	{
		p->moveY = PLAYER_INIT_GRAVITY;

		State = STATE_FALL;
		return;
	}

	//ړ
	MovePlayerRowSideWall();
	MovePlayerJump();

	//XN[␳
	if( ChkScrollLine() == TRUE )
	{
		ScrollLine();
	}

	DrawSprPlayer();
}
//---------------------------------------------------------------------------
void MovePlayerJump()
{
	ST_PLAYER* p = &Player;
	p->y  -= p->moveY;
	p->wy -= p->moveY;

	p->moveY--;
}
//---------------------------------------------------------------------------
void StateFall()
{
	if( ChkFall() == FALSE )
	{
		ST_PLAYER* p = &Player;
		p->moveY = 0;

		State = STATE_MOVE;
		return;
	}

	MovePlayerRowSideWall();
	MovePlayerFall();
	DrawSprPlayer();
}
//---------------------------------------------------------------------------
void MovePlayerFall()
{
	ST_PLAYER* p = &Player;

	u8 reviseY = GetReviseGravity();
	p->y  += reviseY;
	p->wy += reviseY;

	if(p->moveY < PLAYER_MAX_GRAVITY)
	{
		p->moveY++;
	}
}
//---------------------------------------------------------------------------
//̍ہAǂɂ߂荞܂Ȃ悤ɕ␳܂
s16 GetReviseGravity()
{
	ST_PLAYER* p = &Player;

	s16 cnt;
	for(cnt=0; cnt<p->moveY; cnt++)
	{
		if(   ChkTouchWall(p->x, p->wy + cnt + 0, p->cx, p->cy) == FALSE
		   && ChkTouchWall(p->x, p->wy + cnt + 1, p->cx, p->cy) == TRUE )
		{
			return cnt;
		}
	}

	return p->moveY;
}
//---------------------------------------------------------------------------
void GenerateRandomWallLine()
{
	ST_STAGE* p = &Stage;

	s16 x;
	for(x=0; x<STAGE_CX; x++)
	{
		p->bg[x][p->genarateLine] = BG_TILE_EMPTY;
	}

	for(x=0; x<STAGE_CX; x++)
	{
		p->bg[x][p->genarateLine] = BG_TILE_WALL;
	}

	p->genarateLine -= STAGE_GANERATE_INTERVAL;
	if(p->genarateLine < 0)
	{
		p->genarateLine += STAGE_CY - 1;
	}
}
//---------------------------------------------------------------------------
bool ChkScrollLine()
{
	ST_PLAYER* p = &Player;
	return p->y < STAGE_SCROLL_LINE ? TRUE : FALSE;
}
//---------------------------------------------------------------------------
void ScrollLine()
{
	ST_PLAYER* p = &Player;
	ST_STAGE*  s = &Stage;

	//XN[l蒴vZ܂
	u8 scrollY = STAGE_SCROLL_LINE - p->y;

	//ʕ␳ƃvC[̈ړ␳sȂ܂
	s->vofs -= scrollY;
	p->y    += scrollY;

	SetBGScroll();
}
